@@ -1,13 +1,19 @@
-from numbers import Number
-import pandas as pd
-import numpy as np
-from nowcasting_dataset.example import Example, to_numpy
-from nowcasting_dataset import square
-import nowcasting_dataset.time as nd_time
-from dataclasses import dataclass, InitVar
-from typing import List, Tuple, Iterable
-import xarray as xr
+"""  General Data Source Class """
 import itertools
+import logging
+from dataclasses import dataclass, InitVar
+from numbers import Number
+from typing import List, Tuple, Iterable
+
+import numpy as np
+import pandas as pd
+import xarray as xr
+
+import nowcasting_dataset.time as nd_time
+from nowcasting_dataset import square
+from nowcasting_dataset.data_sources.datasource_output import DataSourceOutput
+
+logger = logging.getLogger(__name__)
 
 
 @dataclass
@@ -15,26 +21,47 @@ class DataSource:
     """Abstract base class.
 
     Attributes:
-      history_len: Number of timesteps of history to include in each example.
+      history_minutes: Number of minutes of history to include in each example.
         Does NOT include t0.  That is, if history_len = 0 then the example
         will start at t0.
-      forecast_len: Number of timesteps of forecast to include in each example.
+      forecast_minutes: Number of minutes of forecast to include in each example.
         Does NOT include t0.  If forecast_len = 0 then the example will end
         at t0.  If both history_len and forecast_len are 0, then the example
         will consist of a single timestep at t0.
       convert_to_numpy: Whether or not to convert each example to numpy.
+      sample_period_minutes: The time delta between each data point
     """
-    history_len: int
-    forecast_len: int
+
+    history_minutes: int
+    forecast_minutes: int
     convert_to_numpy: bool
 
     def __post_init__(self):
+        """ Post Init """
+        self.sample_period_minutes = self._get_sample_period_minutes()
+
+        self.history_len = self.history_minutes // self.sample_period_minutes
+        self.forecast_len = self.forecast_minutes // self.sample_period_minutes
+
         assert self.history_len >= 0
         assert self.forecast_len >= 0
+        assert self.history_minutes % self.sample_period_minutes == 0, (
+            f"sample period ({self.sample_period_minutes}) minutes "
+            f"does not fit into historic minutes ({self.forecast_minutes})"
+        )
+        assert self.forecast_minutes % self.sample_period_minutes == 0, (
+            f"sample period ({self.sample_period_minutes}) minutes "
+            f"does not fit into forecast minutes ({self.forecast_minutes})"
+        )
+
         # Plus 1 because neither history_len nor forecast_len include t0.
         self._total_seq_len = self.history_len + self.forecast_len + 1
-        self._history_dur = nd_time.timesteps_to_duration(self.history_len)
-        self._forecast_dur = nd_time.timesteps_to_duration(self.forecast_len)
+        self._history_dur = nd_time.timesteps_to_duration(
+            self.history_len, self.sample_period_minutes
+        )
+        self._forecast_dur = nd_time.timesteps_to_duration(
+            self.forecast_len, self.sample_period_minutes
+        )
 
     def _get_start_dt(self, t0_dt: pd.Timestamp) -> pd.Timestamp:
         return t0_dt - self._history_dur
@@ -43,6 +70,19 @@ class DataSource:
         return t0_dt + self._forecast_dur
 
     # ************* METHODS THAT CAN BE OVERRIDDEN ****************************
+    def _get_sample_period_minutes(self):
+        """
+        This is the default sample period in minutes.
+
+        This functions may be overwritten if
+        the sample period of the data source is not 5 minutes
+        """
+        logging.debug(
+            "Getting sample_period_minutes default of 5 minutes. "
+            "This means the data is spaced 5 minutes apart"
+        )
+        return 5
+
     def open(self):
         """Open the data source, if necessary.
 
@@ -55,23 +95,36 @@ class DataSource:
         pass
 
     def get_batch(
-            self,
-            t0_datetimes: pd.DatetimeIndex,
-            x_locations: Iterable[Number],
-            y_locations: Iterable[Number]) -> List[Example]:
+        self,
+        t0_datetimes: pd.DatetimeIndex,
+        x_locations: Iterable[Number],
+        y_locations: Iterable[Number],
+    ) -> DataSourceOutput:
         """
-        Returns:
-            List of Examples with data converted to Numpy data structures.
+        Get Batch Data
+
+        Args:
+            t0_datetimes: list of timestamps for the datetime of the batches. The batch will also include data
+                for historic and future depending on 'history_minutes' and 'future_minutes'.
+            x_locations: x center batch locations
+            y_locations: y center batch locations
+
+        Returns: Batch data
+
         """
         examples = []
         zipped = zip(t0_datetimes, x_locations, y_locations)
         for t0_datetime, x_location, y_location in zipped:
-            example = self.get_example(t0_datetime, x_location, y_location)
-            if self.convert_to_numpy:
-                example = to_numpy(example)
-            examples.append(example)
+            output: DataSourceOutput = self.get_example(t0_datetime, x_location, y_location)
 
-        return examples
+            if self.convert_to_numpy:
+                output.to_numpy()
+            examples.append(output)
+
+        # could add option here, to save each data source using
+        # 1. # DataSourceOutput.to_xr_dataset() to make it a dataset
+        # 2. DataSourceOutput.save_netcdf(), save to netcdf
+        return DataSourceOutput.create_batch_from_examples(examples)
 
     def datetime_index(self) -> pd.DatetimeIndex:
         """Returns a complete list of all available datetimes."""
@@ -85,8 +138,7 @@ class DataSource:
 
     # ****************** METHODS THAT MUST BE OVERRIDDEN **********************
     def get_locations_for_batch(
-            self,
-            t0_datetimes: pd.DatetimeIndex
+        self, t0_datetimes: pd.DatetimeIndex
     ) -> Tuple[List[Number], List[Number]]:
         """Find a valid geographical location for each t0_datetime.
 
@@ -94,22 +146,18 @@ class DataSource:
             Locations are in OSGB coordinates.
         """
         # TODO: Do this properly, using PV locations!
-        locations = [
-            20_000, 40_000,
-            500_000, 600_000,
-            100_000, 100_000,
-            250_000, 250_000]
+        locations = [20_000, 40_000, 500_000, 600_000, 100_000, 100_000, 250_000, 250_000]
 
         location = np.random.choice(locations, size=(len(t0_datetimes), 2))
 
         return location[:, 0], location[:, 1]
 
     def get_example(
-            self,
-            t0_dt: pd.Timestamp,  #: Datetime of "now": The most recent obs.
-            x_meters_center: Number,  #: Centre, in OSGB coordinates.
-            y_meters_center: Number  #: Centre, in OSGB coordinates.
-    ) -> Example:
+        self,
+        t0_dt: pd.Timestamp,  #: Datetime of "now": The most recent obs.
+        x_meters_center: Number,  #: Centre, in OSGB coordinates.
+        y_meters_center: Number,  #: Centre, in OSGB coordinates.
+    ) -> DataSourceOutput:
         """Must be overridden by child classes."""
         raise NotImplementedError()
 
@@ -117,22 +165,29 @@ class DataSource:
 @dataclass
 class ImageDataSource(DataSource):
     """
+    Image Data source
+
     Args:
       image_size_pixels: Size of the width and height of the image crop
-        returned by get_sample(). """
+        returned by get_sample().
+    """
+
     image_size_pixels: InitVar[int]
     meters_per_pixel: InitVar[int]
 
     def __post_init__(self, image_size_pixels: int, meters_per_pixel: int):
+        """ Post Init """
         super().__post_init__()
         self._square = square.Square(
-            size_pixels=image_size_pixels,
-            meters_per_pixel=meters_per_pixel)
+            size_pixels=image_size_pixels, meters_per_pixel=meters_per_pixel
+        )
 
 
 @dataclass
 class ZarrDataSource(ImageDataSource):
     """
+    A General Zarr Data source
+
     Attributes:
       _data: xr.DataArray data, opened by open().
         x is left-to-right.
@@ -141,58 +196,74 @@ class ZarrDataSource(ImageDataSource):
       consolidated: Whether or not the Zarr store is consolidated.
       channels: The Zarr parameters to load.
     """
+
     channels: Iterable[str]
     #: Mustn't be None, but cannot have a non-default arg in this position :)
     n_timesteps_per_batch: int = None
     consolidated: bool = True
 
     def __post_init__(self, image_size_pixels: int, meters_per_pixel: int):
+        """ Post init """
         super().__post_init__(image_size_pixels, meters_per_pixel)
         self._data = None
         if self.n_timesteps_per_batch is None:
-            raise ValueError('n_timesteps_per_batch must be set!')
+            raise ValueError("n_timesteps_per_batch must be set!")
 
     @property
     def data(self):
+        """ Data property """
         if self._data is None:
-            raise RuntimeError('Please run `open()` before accessing data!')
+            raise RuntimeError("Please run `open()` before accessing data!")
         return self._data
 
     def get_example(
-            self,
-            t0_dt: pd.Timestamp,
-            x_meters_center: Number,
-            y_meters_center: Number
-    ) -> Example:
+        self, t0_dt: pd.Timestamp, x_meters_center: Number, y_meters_center: Number
+    ) -> DataSourceOutput:
+        """
+        Get Example data
+
+        Args:
+            t0_dt: list of timestamps for the datetime of the batches. The batch will also include data
+                for historic and future depending on 'history_minutes' and 'future_minutes'.
+            x_meters_center: x center batch locations
+            y_meters_center: y center batch locations
+
+        Returns: Example Data
+
+        """
         selected_data = self._get_time_slice(t0_dt)
         bounding_box = self._square.bounding_box_centered_on(
-            x_meters_center=x_meters_center, y_meters_center=y_meters_center)
+            x_meters_center=x_meters_center, y_meters_center=y_meters_center
+        )
         selected_data = selected_data.sel(
             x=slice(bounding_box.left, bounding_box.right),
-            y=slice(bounding_box.top, bounding_box.bottom))
+            y=slice(bounding_box.top, bounding_box.bottom),
+        )
 
         # selected_sat_data is likely to have 1 too many pixels in x and y
         # because sel(x=slice(a, b)) is [a, b], not [a, b).  So trim:
         selected_data = selected_data.isel(
-            x=slice(0, self._square.size_pixels),
-            y=slice(0, self._square.size_pixels))
+            x=slice(0, self._square.size_pixels), y=slice(0, self._square.size_pixels)
+        )
 
         selected_data = self._post_process_example(selected_data, t0_dt)
 
         if selected_data.shape != self._shape_of_example:
             raise RuntimeError(
-                'Example is wrong shape! '
-                f'x_meters_center={x_meters_center}\n'
-                f'y_meters_center={y_meters_center}\n'
-                f't0_dt={t0_dt}\n'
-                f'expected shape={self._shape_of_example}\n'
-                f'actual shape   {selected_data.shape}')
+                "Example is wrong shape! "
+                f"x_meters_center={x_meters_center}\n"
+                f"y_meters_center={y_meters_center}\n"
+                f"t0_dt={t0_dt}\n"
+                f"times are {selected_data.time}\n"
+                f"expected shape={self._shape_of_example}\n"
+                f"actual shape {selected_data.shape}"
+            )
 
         return self._put_data_into_example(selected_data)
 
     def geospatial_border(self) -> List[Tuple[Number, Number]]:
-        """Get 'corner' coordinates for a rectangle within the boundary of the
-        data.
+        """
+        Get 'corner' coordinates for a rectangle within the boundary of the data.
 
         Returns List of 2-tuples of the x and y coordinates of each corner,
         in OSGB projection.
@@ -201,28 +272,30 @@ class ZarrDataSource(ImageDataSource):
         data = self._open_data()
         return [
             (data.x.values[x], data.y.values[y])
-            for x, y in itertools.product(
-                [GEO_BORDER, -GEO_BORDER],
-                [GEO_BORDER, -GEO_BORDER])]
+            for x, y in itertools.product([GEO_BORDER, -GEO_BORDER], [GEO_BORDER, -GEO_BORDER])
+        ]
 
     # ****************** METHODS THAT CAN BE OVERRIDDEN **********************
     def _post_process_example(
-            self,
-            selected_data: xr.DataArray,
-            t0_dt: pd.Timestamp) -> xr.DataArray:
+        self, selected_data: xr.DataArray, t0_dt: pd.Timestamp
+    ) -> xr.DataArray:
         return selected_data
 
     # ****************** METHODS THAT MUST BE OVERRIDDEN **********************
     # (in addition to the DataSource methods that must be overridden)
     def open(self) -> None:
-        # We don't want to _open_data() in __init__.
-        # If we did that, then we couldn't copy ZarrDataSource
-        # instances into separate processes.  Instead,
-        # call open() _after_ creating separate processes.
+        """
+        Open the data
+
+        We don't want to _open_data() in __init__.
+        If we did that, then we couldn't copy ZarrDataSource
+        instances into separate processes.  Instead,
+        call open() _after_ creating separate processes.
+        """
         raise NotImplementedError()
 
     def _open_data(self) -> xr.DataArray:
         raise NotImplementedError()
 
-    def _put_data_into_example(self, selected_data: xr.DataArray) -> Example:
+    def _put_data_into_example(self, selected_data: xr.DataArray) -> DataSourceOutput:
         raise NotImplementedError()
